--
-- Calculate CRC16 (CRC-CCITT like in X25, Xmodem, etc) for stream of byte
-- FA: 1996-09-03
--
-- This code is in the public domain:
-- it may be copied, modified and sold freely.
--
--------------------------------------------------------------------------
--
-- A CRC is basically <stream> modulo <divisor>, the remainder of a <stream> divided 
-- by <divisor> in a field where numbers are sets of bits (the whole stream is a huge 
-- binary number) and both addition and substraction are XOR.
--
-- The divisor is a constant binary number which is one bit longer than the divisor.
-- The CRC16 divisor is hex 11021 in our case.
--
-- Is is also called polynomial because some polynomial algebra happens to be 
-- isomorphic with that kind of maths.
--
-- (The CRC16 polynom is x^16+x^12+x^5+x^0 (that is if x=2, 1-0001-0000-0010-0001 
-- or hex 11021, usually advertised as hex 1021 because the last bit is not used 
-- directly during the actual division).
--
--------------------------------------------------------------------------
--
-- IMPORTANT: Note a CRC is designed to dectect errors, that is the same stream 
-- altered with small error-like changes is more unlikely to give a different 
-- CRC than an average checksum would be.
--
-- It is not designed to provide a unique number for a given stream, nor to 
-- provide a number that could not be tampered. That is given a CRC, one could 
-- produce easily something that produces the same CRC. See cryptographic 
-- algorithms such as MD-5 for secure hash values.
--
--------------------------------------------------------------------------
--
-- References:
--
-- Williams, Ross N. "A painless guide to CRC error detection algorithms"
-- http://www.paranoia.com/~filipg/HTML/LINK/F_crc_v3.htm
--
-- Knuth, D.E., "The Art of Computer Programming", Volume 2: 
-- Seminumerical Algorithms, Section 4.6. 
--
-- Sarwate, D.V., "Computation of Cyclic Redundancy Checks via 
-- Table Look-Up", Communications of the ACM, 31(8), pp.1008-1013.
--
-- Ramabadran T.V., Gaitonde S.S., "A tutorial on CRC computations", IEEE 
-- Micro, Aug 1988.
--

indexing
	description: "Calculate CRC-16 using CCITT polynom x^16+x^12+x^5+x^0"
	compiler: "This class uses basic types proprietary to TowerEiffel"

class CRC16 

inherit ANY redefine is_equal end
	
creation make

feature { CRC16 }
	
	crc16 : BIT 16
			-- Current CRC value.

feature -- General

	is_equal(other : like Current) : BOOLEAN is
			-- Equality.
		do
			Result := crc16 = other.crc16
		end

feature
	
	make,reset is
			-- Initialise CRC.
		do
			crc16 := 0B --Feed16
		end

feature -- Status

	value : INTEGER is
			-- CRC value.
		do
			Result := crc16.to_uinteger_16.to_integer
		end

feature -- Change
	
	append_byte(input_byte : INTEGER) is
		require
			byte: input_byte >= 0 and input_byte < 256
		local
			byte_16 : BIT 16
		do
			-- Here is the classic black magic.
			
			-- It is a short cut for doing a CRC-division at the bit level, but 
			-- with 8 bits done at a time using a precomputed table of all possible 
			-- values for a byte (the divisor xored at various shifted position).

			-- To confuse the matter further, and avoid adding the required 16 
			-- zero bits at the end, the input bytes are fed directly through 
			-- the table.

			byte_16 := input_byte.to_uinteger_16.to_bit_16
			crc16 := crc16_table.item(
						(((crc16 ^ 8) xor byte_16) and 011111111B).to_uinteger_16.to_integer
							-- top_byte(crc16) BITOR input_byte
						)
							xor 
						(crc16 ^ -8) -- crc16 << 8
		end

	
feature { NONE } -- Constant

	crc16_table : ARRAY[BIT 16] is
			-- Magic CRC16 table (preprocessed xor mod 2 division for every value of a byte).
		local
			i, bits : INTEGER
			temp,res : BIT 16
		once
			!!Result.make(0,255)
			from i := 0 variant 256 - i until i > 255 loop
				
				res := 0B
				temp := i.to_uinteger_16.to_bit_16 ^ -8
				
				from bits := 1 variant 9 - bits until bits > 8 loop
					if ((temp xor res) and 1000000000000000B) /= 0B then
						res := (res ^ -1) xor 01000000100001B -- 0x1021 -- res << 1 ^ divisor
					else
						res := res ^ -1 -- res << 1 
					end
					temp := temp ^ -1
					bits := bits + 1
				end
				
				Result.put(res,i)
				
				i := i + 1
			end
		ensure
			done: Result.lower = 0 and Result.upper = 255 -- 0x00 to 0xFF
		end

feature -- Helper
	
	append_string(str : STRING) is
			-- Append string's characters to the CRC.
		require
			portability: -- sizeof(CHARACTER) = 8 bits.
			ok: str /= Void
		local
			i, cnt : INTEGER
		do
			cnt := str.count
			from i := 1 variant cnt + 1 - i until i > cnt loop
				append_byte(str.item(i).code) --? unicode
				i := i + 1
			end
		end

	append_integer(int : INTEGER) is
			-- Append a 32 bit integer.
		require
			portability: -- sizeof(INTEGER) = 32 bits.
		local
			i : INTEGER
		do
			from i := 24 variant i // 8 + 1 until i < 0 loop
				append_byte((int // 2 ^ i).to_uinteger_8.to_integer) --?
				i := i - 8
			end
		end

end -- class CRC16

